package	jp.co.sds_corp.util.daemon.core;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.daemon.filter.IFilter;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

/**
* TimerDaemon の実装です。
* 指定された間隔でサービスを呼び出します。
<P>
XMLは以下の形式です。
<PRE>
&lt;Daemon name="STP" class="jp.co.sds_corp.util.daemon.core.TimerDaemon"&gt;
	&lt;delay type="sec"&gt;0&lt;/delay&gt;
	&lt;interval type="sec"&gt;10&lt;/interval&gt;
	&lt;Filter class="jp.co.sds_corp.util.daemon.filter.TimeRange"
		type="allow" from="06:00:00" to="21:00:00"/&gt;
	&lt;Service name="SVC" class="service.SVC"&gt;
		...
	&lt;/Service&gt;
&lt;/Daemon&gt;
</PRE>
delay要素で指定された秒数後に、interval要素で指定された秒数間隔でサービスを
起動します。
また、Filter要素で制限された時間内のみ、サービスを起動します。
{@link IFilter}で許可と拒否が重なった時間では、処理を起動しません。
*/
public class TimerDaemon extends AbstractDaemon
		implements TimerDaemonMBean
{
	CmnLogger logger = CmnLogFactory.getInstance().getLogger(this);
	private transient ServiceTask task;
	private transient Timer timer = null;
	Vector<Interval> vecItv = new Vector<Interval>();
	Map<Interval,ServiceTask> mapTask = new HashMap<Interval,ServiceTask>();
	private long lDelay = 0;
	private boolean bFixed = false;

	/**
	* 設定情報でこのデーモンを初期化します。
	*/
	public void init( IConfiguration config )
		throws Exception
	{
		super.init( config );
		IConfiguration c;
		Iterator itr = config.getChilds( "interval" );
		while ( itr.hasNext()) {
			c = (IConfiguration )itr.next();
			Interval itv = new Interval();
			itv.init( c );
			if ( itv.getMilliSec() < 0 ) {
				throw( new IllegalArgumentException( "Negative interval."
						+ "[" + itv + "]" ));
			}
			vecItv.add( itv );
		}
		c = config.getChild( "delay" );
		if ( c != null ) {
			Interval delay = new Interval();
			delay.init( c );
			lDelay = delay.getMilliSec();
			if ( lDelay < 0 ) {
				lDelay = 0;
			}
		}

		String var;
		var = config.get( "fixed", null );
		if ( var != null && var.equals( "true" )) {
			bFixed = true;
		}
	}

	/**
	* 遅延時間を設定します。
	* @param	var		遅延時間（秒）
	*/
	public void setDelay( long var ) {
		this.lDelay = var;
	}
	/**
	* 遅延時間を取得します。
	* @return	遅延時間（秒）
	*/
	public long getDelay() {
		return lDelay;
	}

	/**
	* 固定間隔であるかどうかを設定します。
	* @param	var		true:固定間隔 
	*/
	public void setFixed( boolean var ) {
		this.bFixed = var;
	}
	/**
	* 固定間隔であるかどうかを取得します。
	* @return	true:固定間隔 
	*/
	public boolean isFixed() {
		return bFixed;
	}

	/**
	* デーモンを開始します。
	*/
	public void start() {
		if ( getStatus() == Status.ERROR ) {
			return;
		}
		logger.debug( "start:" + getName() + ":" + getClass().getName()
							+ " delay=" + lDelay + " fix=" + bFixed );
		startTimer();
		setStatus( Status.ALIVE );
	}

	/**
	* タイマを起動します。
	* 現在起動中のタイマはキャンセルされます。
	*/
	public void startTimer() {
		stopTimer();
		timer = new Timer();
		for ( int i = 0; i < vecItv.size(); i++ ) {
			Interval itv = (Interval )vecItv.get( i );
			startTimer( itv );
		}
	}

	public void startTimer( Interval itv ) {
		task = new ServiceTask();
		mapTask.put( itv, task );
		if ( bFixed ) {
			timer.scheduleAtFixedRate( task, lDelay, itv.getMilliSec() );
		} else {
			timer.schedule( task, lDelay, itv.getMilliSec() );
		}
	}

	/**
	* タイマを停止します。
	*/
	public void stopTimer() {
		if ( timer != null ) {
			timer.cancel();
			for ( ServiceTask task : mapTask.values() ) {
				task.cancel();
			}
			mapTask.clear();
			task.cancel();
		}
	}

	/**
	* TimerDaemonに間隔を設定します。
	* @param	itv		間隔
	*/
	public void setInterval( Interval itv ) throws Exception {
		Interval org = vecItv.set( 0, itv );
		if ( org != null ) {
			removeInterval( org );
		}
		startTimer( itv );
	}
	/**
	* TimerDaemonの間隔を取得します。
	* @return	間隔
	*/
	public Interval getInterval() {
		Interval itv = (Interval )vecItv.get( 0 );
		return itv;
	}
	/**
	* TimerDaemonの間隔を削除します。
	* @return	削除されたスケジュール
	*/
	public Interval removeInterval() throws Exception {
		return removeInterval( 0 );
	}

	/**
	* TimerDaemonに間隔を追加します。
	* @param	itv		間隔
	*/
	public void addInterval( Interval itv ) throws Exception {
		vecItv.add( itv );
		startTimer( itv );
	}

	/**
	* TimerDaemonの間隔を削除します。
	* @param	idx		削除するスケジュールのインデックス
	* @return	削除されたスケジュール
	*/
	public Interval removeInterval( int idx ) throws Exception {
		Interval itv = (Interval )vecItv.remove( idx );
		return removeInterval( itv );
	}
	/**
	* TimerDaemonの間隔を削除します。
	* @param	itv		削除するスケジュール
	* @return	削除されたスケジュール
	*/
	public Interval removeInterval( Interval itv ) throws Exception {
		ServiceTask task = (ServiceTask )mapTask.remove( itv );
		if ( task != null ) {
			task.cancel();
		}
		return itv;
	}

	/**
	* 起動間隔の一覧を取得します。
	* @return	起動間隔の一覧
	*/
	public List<Interval> listInterval() {
		return vecItv;
	}

	/**
	* 直ちに処理を実行します。
	*/
	public void startImmediate() throws Exception {
		ServiceTask task = new ServiceTask();
		timer.schedule( task, 0 );
	}

	/**
	* デーモンを停止します。
	*/
	public void terminate() {
		logger.debug( "terminate:" + getName() + ":" + getClass().getName() );
		stopTimer();
		getService().stop();
		setStatus( Status.DEAD );
	}

	/**
	* デーモンの全ての Thread の終了を待ちます。
	*/
	public void joinAll() {
	}

	/**
	* このデーモンの文字列表現を返します。
	* @return	このデーモンの文字列表現
	*/
	public String toString() {
		String ret = super.toString() + ";delay=" + lDelay;
		for ( int i = 0; i < vecItv.size(); i++ ) {
			Interval itv = (Interval )vecItv.get( i );
			ret += "\n\t" + itv.toString();
		}
		return ret;
	}

	/**
	* 設定されている内容をXML形式で出力します。
	* @param	config	設定情報
	*/
	public void saveConfig( IConfiguration config ) {
		super.saveConfig( config );
		IConfiguration c;
		Interval delay = new Interval();
		delay.setType( Interval.Type.MILLISEC );

		c = config.addChild( "delay" );
		delay.setMilliSec( lDelay );
		delay.saveConfig( c );
		for ( int i = 0; i < vecItv.size(); i++ ) {
			Interval itv = (Interval )vecItv.get( i );
			c = config.addChild( "interval" );
			itv.saveConfig( c );
		}
		c = config.addChild( "fixed", String.valueOf( bFixed ) );
	}

	/**
	* サービスを起動するTimerTask
	*/
	private class ServiceTask extends TimerTask implements Serializable
	{
		public void run() {
			if ( !isValid()) {
				return;
			}
			if ( isDebug()) {
				logger.debug( "start TIMER service:" + getService() );
			}
			setStatus( IDaemon.Status.RUN );
			try {
				Thread.currentThread().setName( TimerDaemon.this.getName()
						+ ":" + TimerDaemon.this.getService().getName() );
				getService().start();
			} catch ( Throwable t ) {
				logger.error( "Service error", t );
				setStatus( IDaemon.Status.ERROR, t.getMessage() );
			}
			setStatus( IDaemon.Status.ALIVE );
			if ( isDebug()) {
				logger.debug( "end TIMER service:" + getService() );
			}
		}
	}
}

