package	jp.co.sds_corp.util.daemon.db;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.sql.Types;
import java.util.Iterator;
import java.util.Vector;

import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.daemon.AbstractService;
import jp.co.sds_corp.util.db.DbAccess;
import jp.co.sds_corp.util.db.DbColumn;
import jp.co.sds_corp.util.db.DbInfo;
import jp.co.sds_corp.util.db.DbTable;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

/**
* テーブルの内容をテキストファイルに出力します。
<P>
XMLは以下の形式です。
<PRE>
&lt;Service name="Test" class="jp.co.sds_corp.util.daemon.db.Exporter"
	outputdir="tmp" ext="bcp" colseparator="\t" rowseparator="\n"&gt;
	&lt;db type="properties"/&gt;
	&lt;catalogsql&gt;select TABLE_NAME from USER_TABLES where DROPPED='NO'&lt;/catalogsql&gt;
	&lt;table&gt;TABLEA&lt;/tabl&gt;
	&lt;table&gt;TABLEB&lt;/tabl&gt;
&lt;/Service&gt;
</PRE>

outputdir属性で、出力ファイルのディレクトリを指定します。
デフォルトは"."です。<br>
ext属性で、出力ファイルの拡張子を指定します。
デフォルトは"bcp"です。<br>
colseparator属性で、出力ファイルのカラムのセパレータを指定します。
デフォルトは"\t"です。<br>
rowseparator属性で、出力ファイルの行のセパレータを指定します。
デフォルトは"\n"です。<br>

<P>
db要素は以下の子要素のいずれかを持つことができます。
詳細は{@link DbInfo},{@link DbAccess}を参照してください。

dbキーを指定する場合
<PRE>
	&lt;db type="properties"&gt;
		&lt;key&gt;db_key&lt;/key&gt;
	&lt;/db&gt;
</PRE>

JNDIを使用する場合
<PRE>
	&lt;db&gt;
	&lt;db type="jndi"&gt;
		&lt;context&gt;org.jnp.interfaces.NamingContextFactory&lt;/context&gt;
		&lt;url&gt;localhost&lt;/url&gt;
		&lt;datasource&gt;java:/DefaultDS&lt;/datasource&gt;
	&lt;/db&gt;
</PRE>

JDBCドライバを直接使用する場合
<PRE>
	&lt;db type="jdbc"&gt;
		&lt;driver&gt;com.sybase.jdbc2.jdbc.SybDriver&lt;/driver&gt;
		&lt;url&gt;jdbc:sybase:Tds:localhost:4100/db&lt;/url&gt;
		&lt;user&gt;user&lt;/user&gt;
		&lt;password&gt;password&lt;/password&gt;
	&lt;/db&gt;
</PRE>
<P>
table要素で、保存するテーブル名を指定します。
catalogsql要素を実行して、結果セットの1番目のカラムをテーブル名として使用します。
catalog要素、table
<P>
*/
public class Exporter extends AbstractService
		implements ExporterMBean
{
	CmnLogger logger = CmnLogFactory.getInstance().getLogger(this);
	private String		rowsep = "\n";
	private String		colsep = "\t";
	private String		dateformat = "yyyy-MM-dd HH:mm:ss";
	private String		timeformat = "HH:mm:ss";
	private String		floatformat = "0.#";
	private String		ext;
	private String		dir;
	private Vector<String>		vecTbl = new Vector<String>();
	private Vector<String>		vecSql = new Vector<String>();
	private	DbAccess	db;
	private IConfiguration config;
	private IConfiguration dbconf;

	/**
	* このクラスを設定情報で初期化します。
	* @param	config	設定情報
	*/
	public void init( IConfiguration config ) throws Exception {
		this.config = config;
		dir = config.getAttributes().get( "outputdir", "." );
		ext = config.getAttributes().get( "ext", "bcp" );
		dbconf = config.getChild( "db" );
		db = new DbAccess( dbconf );

		Iterator<IConfiguration> itr;
		itr = config.getChilds( "catalogsql" );
		while ( itr.hasNext()) {
			IConfiguration c = (IConfiguration )itr.next();
			vecSql.add( c.get() );
		}
		itr = config.getChilds( "table" );
		while ( itr.hasNext()) {
			IConfiguration c = (IConfiguration )itr.next();
			vecTbl.add( c.get() );
		}
	}

	/**
	* 出力するテーブルの一覧を取得します。
	* @param	config	設定情報
	* @param	db		DbAccess
	* @param	vec		テーブル一覧
	*/
	protected void listTable( IConfiguration config, DbAccess db,
			Vector<String> vec )
		throws Exception
	{
		vec.clear();
		for ( int i = 0; i < vecSql.size(); i++ ) {
			String sql = (String )vecSql.get( i );
			selectCatalog( db, vec, sql );
		}
		for ( int i = 0; i < vecTbl.size(); i++ ) {
			String tbl = (String )vecTbl.get( i );
			vec.add( tbl );
		}
	}

	/**
	* SQLを実行してテーブル名を取得します。
	* @param	db		DbAccess
	* @param	vec		テーブル一覧
	* @param	sql		実行するSQL
	*/
	protected void selectCatalog( DbAccess db, Vector<String> vec, String sql )
		throws Exception
	{
		DbTable tbl = new DbTable();
		db.executeQuery( sql );
		for ( ;; ) {
			if ( !db.fetch( tbl )) {
				break;
			}
			vec.add( tbl.getString( 0 ));
		}
	}

	/**
	* テーブルをエクスポートします。
	*/
	public void start() {
		setStatus( Status.RUN );
		try {
			db.connect();
			Vector<String> vec = new Vector<String>();
			listTable( config, db, vec );
			for ( int i = 0; i < vec.size(); i++ ) {
				String tbl = (String )vec.get( i );
				export( db, tbl );
			}
			setStatus( Status.WAIT );
		} catch ( Exception ex ) {
			setStatus( Status.ERROR );
			logger.fatal( "table export error", ex );
		} finally {
			try {
				db.disconnect();
			} catch ( Exception ee ) {
			}
		}
	}

	/**
	* テーブルをファイルに出力します。
	* @param	db		DbAccess
	* @param	table	テーブル名
	*/
	protected void export( DbAccess db, String table )
		throws Exception
	{
		String fn = dir + File.separator + table + "." + ext;
		PrintWriter pw =
			new PrintWriter( new BufferedWriter( new FileWriter( fn )));
		String sql = "select * from " + table;
		DbTable tbl = new DbTable();
		db.executeQuery( sql );
		int nRow = 0;
		for ( ;; ) {
			if ( !db.fetch( tbl )) {
				break;
			}
			for ( int i = 0; i < tbl.getColumnCount(); i++ ) {
				String data = getColumnString( tbl, i );
				pw.print( data );
				if ( i == tbl.getColumnCount() - 1 ) {
					pw.print( rowsep );
				} else {
					pw.print( colsep );
				}
			}
			nRow++;
		}
		pw.flush();
		pw.close();
		logger.debug( fn + " " + nRow + " rows." );
	}

	/**
	* カラムのデータを文字列に変換します。
	* @param	tbl		テーブル情報
	* @param	nCol	カラム番号
	*/
	protected String getColumnString( DbTable tbl, int nCol ) {
		DbColumn col = tbl.getColumn( nCol );
		String ret;
		switch ( col.getType()) {
		case Types.DATE:
		case Types.TIMESTAMP:
			try {
				ret = tbl.formatDate( col.getName(), getDateFormat() );
			} catch ( Exception ex ) {
				ret = "";
			}
			break;
		case Types.TIME:
			try {
				ret = tbl.formatDate( col.getName(), getTimeFormat() );
			} catch ( Exception ex ) {
				ret = "";
			}
			break;
		case Types.FLOAT:
		case Types.DOUBLE:
			ret = tbl.formatDecimal( col.getName(), getFloatFormat() );
			break;
		default:
			ret = tbl.getString( nCol );
			break;
		}
		if ( ret == null ) {
			ret = "";
		}
		return ret;
	}

	/**
	* 処理を終了します。
	*/
	public void stop() {
	}

	/**
	* ファイルの保存先を取得します。
	*/
	public String getOutputDir() {
		return this.dir;
	}
	/**
	* ファイルの保存先を設定します。
	* @param	var		ファイルの保存先
	*/
	public void setOutputDir( String var ) {
		this.dir = var;
	}

	/**
	* 日時の出力フォーマットを取得します。
	*/
	public String getDateFormat() {
		return this.dateformat;
	}
	/**
	* 日時の出力フォーマットを設定します。
	* @param	var		フォーマット
	*/
	public void setDateFormat( String var ) {
		this.dateformat = var;
	}

	/**
	* 時間の出力フォーマットを取得します。
	*/
	public String getTimeFormat() {
		return this.timeformat;
	}
	/**
	* 時間の出力フォーマットを設定します。
	* @param	var		フォーマット
	*/
	public void setTimeFormat( String var ) {
		this.timeformat = var;
	}

	/**
	* floatの出力フォーマットを取得します。
	*/
	public String getFloatFormat() {
		return this.floatformat;
	}
	/**
	* floatの出力フォーマットを設定します。
	* @param	var		フォーマット
	*/
	public void setFloatFormat( String var ) {
		this.floatformat = var;
	}

	/**
	* 設定されている内容をXML形式で出力します。
	* @param	config	設定情報
	*/
	public void saveConfig( IConfiguration config ) {
		super.saveConfig( config );
		config.setAttribute( "outputdir", dir );
		config.setAttribute( "ext", ext );
		config.addChild( dbconf );
		for ( int i = 0; i < vecSql.size(); i++ ) {
			String sql = (String )vecSql.get( i );
			config.addChild( "catalogsql", sql );
		}
		for ( int i = 0; i < vecTbl.size(); i++ ) {
			String tbl = (String )vecTbl.get( i );
			config.addChild( "table", tbl );
		}
	}

	/**
	* このサービスの文字列表現を返します。
	*/
	public String toString() {
		String ret = super.toString();
		if ( vecSql.size() > 0 ) {
			ret += "\n\tsql[";
			for ( int i = 0; i < vecSql.size(); i++ ) {
				if ( i > 0 ) {
					ret += ",";
				}
				ret += (String )vecSql.get( i );
			}
			ret += "]";
		}
		if ( vecTbl.size() > 0 ) {
			if ( vecSql.size() > 0 ) {
				ret += "\n";
			}
			ret += "\ttable[";
			for ( int i = 0; i < vecTbl.size(); i++ ) {
				if ( i > 0 ) {
					ret += ",";
				}
				ret += (String )vecTbl.get( i );
			}
			ret += "]";
		}
		return ret;
	}
}
