package	jp.co.sds_corp.util.daemon.remote.jms;

import javax.jms.JMSException;

import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.daemon.remote.AbstractServer;
import jp.co.sds_corp.util.daemon.remote.Commander;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;
import jp.co.sds_corp.util.msg.IMessage;
import jp.co.sds_corp.util.msg.IMessageAdapter;
import jp.co.sds_corp.util.msg.IReceiveListener;
import jp.co.sds_corp.util.msg.MessageAdapterFactory;

/**
* デーモンマネージャのRMI I/F のサーバです。
<pre>
&lt;Remote type="JMS"&gt;
	&lt;Type&gt;jmsTopic&lt;/Type&gt;
	&lt;InitialContext&gt;weblogic.jndi.WLInitialContextFactory&lt;/InitialContext&gt;
	&lt;ConnectionFactory&gt;jms/queue/ConnectionFactory&lt;/ConnectionFactory&gt;
	&lt;URL&gt;http://localhost:8080&lt;/URL&gt;
	&lt;User&gt;user&lt;/User&gt;
	&lt;Password&gt;passwd&lt;/Password&gt;
	&lt;Receive&gt;jms/queue/request&lt;/Receive&gt;
	&lt;Send&gt;jms/queue/response&lt;/Send&gt;
&lt;/Remote&gt;
</pre>
*/
public class JMSServer extends AbstractServer
{
	CmnLogger logger = CmnLogFactory.getInstance().getLogger(this);
	IMessageAdapter	adapter;
	Commander	cmd;

	/**
	* 設定情報で初期化します。
	* @param	config	設定情報
	*/
	public void init( IConfiguration config ) throws Exception {
		super.init( config );
		setType( "JMS" );
		logger.debug( "start JMSServer" );
		MessageAdapterFactory factory = new MessageAdapterFactory();
		adapter = factory.connect( config );
		logger.debug( "start JMSServer " + adapter );

		adapter.createReceiver( config.get( "Receive" ), null, false );
		adapter.addReceiveListener( new Receiver() );
		adapter.createSender( config.get( "Send" ));
		adapter.start();

		cmd = new Commander( getManager() );
	}

	/**
	* RMIサーバを終了します。
	*/
	public void exit() {
		try {
			logger.debug( "stop JMSServer " + adapter );
			adapter.stop();
		} catch ( JMSException je ) {
			Exception le = je.getLinkedException();
			if ( le != null ) {
				logger.error( "linked exception:", le );
			}
			logger.error( "stop error", je );
		} catch ( Exception ex ) {
		}
	}

	private class Receiver implements IReceiveListener {
		/**
		* JMSメッセージを受信した時に呼び出されます。
		* @param    mAdapter	受信したメッセージアダプタ
		* @param    msg			受信したメッセージ
		* @param    type		メッセージのタイプ
		*/
		public void onMessage( IMessageAdapter mAdapter, IMessage msg )
				
		{
			// received( msg );
			ReceiveThread thr = new ReceiveThread( msg );
			thr.start();
		}
	}

	protected class ReceiveThread extends Thread {
		IMessage msg;
		ReceiveThread( IMessage msg ) {
			this.msg = msg;
			setName( "JMSServer:ReceiveThread" );
		}
		public void run() {
			received( msg );
		}
	}

	protected void received( IMessage msg ) {
		Object ret = command( msg );
		try {
			IMessage res = adapter.copyMessage( msg );
			res.setObject( "response", ret );
			adapter.writeMessage( res, 5000 );
		} catch ( JMSException je ) {
			Exception le = je.getLinkedException();
			if ( le != null ) {
				logger.error( "linked exception:", le );
			}
			logger.error( "send error", je );
		} catch ( Exception ex ) {
			logger.error( "send error", ex );
		}
	}

	protected Object command( IMessage msg ) {
		Object ret = null;
		try {
			String method = (String )msg.getObject( "command" );
			Class<?>[] cls = (Class<?>[] )msg.getObject( "class" );
			Object[] prm = (Object[] )msg.getObject( "param" );
			ret = super.command( cmd, method, cls, prm );
		} catch ( Exception ex ) {
			if ( isDebug()) {
				logger.warn( "command invoke error", ex );
			}
			ret = ex;
		}
		return ret;
	}

}
